import re as regex

from elody.policies.permission_handler import get_permissions
from flask import Request  # pyright: ignore
from inuits_policy_based_auth import BaseAuthorizationPolicy  # pyright: ignore


class TenantRequestPolicy(BaseAuthorizationPolicy):
    def authorize(self, policy_context, user_context, request_context):
        request: Request = request_context.http_request
        if not regex.match("^(/[^/]+/v[0-9]+)?/tenants$", request.path):
            return policy_context

        set_restricting_filter = True
        for role in user_context.x_tenant.roles:
            permissions = get_permissions(role, user_context)
            if "tenant" in permissions.get("read", {}).keys():
                set_restricting_filter = False
                break

        if set_restricting_filter:
            user_context.access_restrictions.filters = {
                "relations": {
                    "$elemMatch": {
                        "key": {"$in": user_context.bag["user_ids"]},
                        "type": "isTenantFor",
                    }
                }
            }

        policy_context.access_verdict = True
        return policy_context



from elody.object_configurations.elody_configuration import (
    ElodyConfiguration,
)
from elody.util import send_cloudevent
from os import getenv


class JobConfiguration(ElodyConfiguration):
    SCHEMA_TYPE = "elody"
    SCHEMA_VERSION = 1

    def crud(self):
        crud = {"collection": "jobs", "collection_history": ""}
        return {**super().crud(), **crud}

    def document_info(self):
        return super().document_info()

    def logging(self, flat_item, **kwargs):
        return super().logging(flat_item, **kwargs)

    def migration(self):
        return super().migration()

    def serialization(self, from_format, to_format):
        return super().serialization(from_format, to_format)

    def validation(self):
        return super().validation()

    def _post_crud_hook(self, *, crud, document, get_rabbit, **kwargs):
        if crud == "create":
            send_cloudevent(
                get_rabbit(),
                getenv("MQ_EXCHANGE", "dams"),
                "dams.job_created",
                document,
            )
            if parent_id := kwargs.get("parent_id"):
                send_cloudevent(
                    get_rabbit(),
                    getenv("MQ_EXCHANGE", "dams"),
                    "dams.job_changed",
                    {
                        "id": parent_id,
                        "patch": {
                            "relations": [
                                {"key": document["_id"], "type": "isParentJobOf"}
                            ]
                        },
                    },
                )
        elif crud == "update":
            send_cloudevent(
                get_rabbit(),
                getenv("MQ_EXCHANGE", "dams"),
                "dams.job_changed",
                document,
            )



import elody.util as util
import json
import os

from apscheduler.triggers.cron import CronTrigger
from importlib import import_module
from inuits_policy_based_auth.exceptions import (
    PolicyFactoryException,
)


def load_apps(flask_app, logger):
    apps = util.read_json_as_dict(os.getenv("APPS_MANIFEST"), logger)
    for app in apps:
        for resource in apps[app].get("resources", []):
            api_bp = import_module(f"apps.{app}.resources.{resource}").api_bp
            flask_app.register_blueprint(api_bp)


def load_jobs(scheduler, logger):
    apps = util.read_json_as_dict(os.getenv("APPS_MANIFEST"), logger)
    for app in apps:
        for job, job_properties in apps[app].get("jobs", {}).items():
            try:
                job_class = __get_class_from_module(
                    import_module(f"apps.{app}.cron_jobs.{job}")
                )
                scheduler.add_job(
                    job_class(),
                    CronTrigger.from_crontab(
                        job_properties.get("expression", "0 0 * * *")
                    ),
                )
            except ModuleNotFoundError:
                pass


def load_policies(policy_factory, logger, permissions={}):
    if permissions:
        from elody.policies.permission_handler import set_permissions

        set_permissions(permissions)

    apps = util.read_json_as_dict(os.getenv("APPS_MANIFEST", ""), logger)
    for app in apps:
        try:
            auth_type = "authentication"
            for policy_module_name in apps[app]["policies"].get(auth_type):
                policy = __get_class(app, auth_type, policy_module_name)
                policy = __instantiate_authentication_policy(
                    policy_module_name, policy, logger
                )
                policy_factory.register_authentication_policy(f"apps.{app}", policy)
            auth_type = "authorization"
            for policy_module_name in apps[app]["policies"].get(auth_type):
                policy = __get_class(app, auth_type, policy_module_name)
                policy_factory.register_authorization_policy(f"apps.{app}", policy())
            # FIXME: don't always set last app as fallback
            policy_factory.set_fallback_key_for_policy_mapping(f"apps.{app}")
        except Exception as error:
            raise PolicyFactoryException(
                f"Policy factory was not configured correctly: {str(error)}"
            ).with_traceback(error.__traceback__)


def load_queues(logger):
    import_module("resources.queues")
    apps = util.read_json_as_dict(os.getenv("APPS_MANIFEST"), logger)
    for app in apps:
        try:
            import_module(f"apps.{app}.resources.queues")
        except ModuleNotFoundError:
            pass


def __get_class(app, auth_type, policy_module_name):
    locations = [
        policy_module_name,
        f"apps.{app}.policies.{auth_type}.{policy_module_name}",
        f"elody.policies.{auth_type}.{policy_module_name}",
        f"inuits_policy_based_auth.{auth_type}.policies.{policy_module_name}",
    ]
    for location in locations:
        try:
            module = import_module(location)
            break
        except ModuleNotFoundError:
            pass
    else:
        raise ModuleNotFoundError(f"Policy {policy_module_name} not found")
    policy = __get_class_from_module(module)
    return policy


def __get_class_from_module(module):
    class_name = module.__name__.split(".")[-1].title().replace("_", "")
    return getattr(module, class_name)


def __instantiate_authentication_policy(policy_module_name, policy, logger):
    allow_anonymous_users = os.getenv("ALLOW_ANONYMOUS_USERS", False) in [
        "True",
        "true",
        True,
    ]
    if policy_module_name == "token_based_policies.authlib_flask_oauth2_policy":
        token_schema = __load_token_schema()
        allowed_issuers = os.getenv("ALLOWED_ISSUERS")
        return policy(
            logger,
            token_schema,
            os.getenv("STATIC_ISSUER"),
            os.getenv("STATIC_PUBLIC_KEY"),
            allowed_issuers.split(",") if allowed_issuers else None,
            allow_anonymous_users,
        )
    if policy_module_name == "token_based_policies.tenant_token_roles_policy":
        token_schema = __load_token_schema()
        return policy(
            token_schema,
            os.getenv("ROLE_SCOPE_MAPPING", "role_scope_mapping.json"),
            allow_anonymous_users,
        )
    if policy_module_name == "elody.policies.authentication.multi_tenant_policy":
        tenant_defining_types = os.getenv("TENANT_DEFINING_TYPES")
        tenant_defining_types = (
            tenant_defining_types.split(",") if tenant_defining_types else []
        )
        return policy(
            os.getenv("TENANT_DEFINING_HEADER", "X-tenant-id"),
            tenant_defining_types,
            os.getenv("AUTO_CREATE_TENANTS", False) in ["True", "true", True],
        )
    return policy()


def __load_token_schema() -> dict:
    token_schema_path = os.getenv("TOKEN_SCHEMA", "token_schema.json")
    with open(token_schema_path, "r") as token_schema:
        return json.load(token_schema)



class BaseObjectMigrator:
    def __init__(self, *, status, silent=False):
        self._status = status
        self._silent = silent

    @property
    def status(self):
        return self._status

    @property
    def silent(self):
        return self._silent

    def bulk_migrate(self, *, dry_run=False):  # pyright: ignore
        pass

    def lazy_migrate(self, item, *, dry_run=False):  # pyright: ignore
        return item



import re as regex

from elody.policies.helpers import get_content
from configuration import get_object_configuration_mapper  # pyright: ignore
from elody.policies.permission_handler import (
    get_permissions,
    handle_single_item_request,
    mask_protected_content_post_request_hook,
)
from elody.util import interpret_flat_key
from flask import Request  # pyright: ignore
from inuits_policy_based_auth import BaseAuthorizationPolicy  # pyright: ignore
from inuits_policy_based_auth.contexts.policy_context import (  # pyright: ignore
    PolicyContext,
)
from inuits_policy_based_auth.contexts.user_context import (  # pyright: ignore
    UserContext,
)


class GenericObjectRequestPolicy(BaseAuthorizationPolicy):
    def authorize(
        self, policy_context: PolicyContext, user_context: UserContext, request_context
    ):
        request: Request = request_context.http_request
        if not regex.match("^(/[^/]+/v[0-9]+)?/[^/]+$", request.path):
            return policy_context

        for role in user_context.x_tenant.roles:
            permissions = get_permissions(role, user_context)
            if not permissions:
                continue

            rules = [PostRequestRules, GetRequestRules]
            access_verdict = None
            for rule in rules:
                access_verdict = rule().apply(user_context, request, permissions)
                if access_verdict != None:
                    policy_context.access_verdict = access_verdict
                    if not policy_context.access_verdict:
                        return policy_context

            if policy_context.access_verdict:
                return policy_context

        return policy_context


class PostRequestRules:
    def apply(
        self, user_context: UserContext, request: Request, permissions
    ) -> bool | None:
        if request.method != "POST":
            return None
        if request.args.get("dry_run", False):
            return True
        if regex.match(r"^/batch?$", request.path):
            return True

        content = get_content(request.json, request, request.json)
        return handle_single_item_request(
            user_context, request.json, permissions, "create", content
        )


class GetRequestRules:
    def apply(
        self, user_context: UserContext, request: Request, permissions
    ) -> bool | None:
        if request.method != "GET":
            return None
        type_query_parameter = (
            "mediafile"
            if regex.match(r"^/mediafiles(?:\?(.*))?$", request.path)
            else request.args.get("type")
        )
        allowed_item_types = list(permissions["read"].keys())
        filters = []

        if type_query_parameter:
            if type_query_parameter in allowed_item_types:
                config = get_object_configuration_mapper().get(type_query_parameter)
                object_lists = config.document_info()["object_lists"]

                restrictions = permissions["read"][type_query_parameter].get(
                    "object_restrictions", {}
                )
                for key, value in restrictions.items():
                    keys_info = interpret_flat_key(key, object_lists)
                    filters.append(
                        _build_nested_matcher(object_lists, keys_info, value)
                    )
            else:
                return None
        else:
            filters = [
                {"type": {"$in": allowed_item_types}},
                {
                    "relations": {
                        "$elemMatch": {
                            "key": user_context.bag.get(
                                "tenant_defining_entity_id", user_context.x_tenant.id
                            ),
                            "type": [
                                user_context.bag["tenant_relation_type"],
                                "belongsTo",
                            ],
                        }
                    }
                },
            ]

        user_context.access_restrictions.filters = filters
        user_context.access_restrictions.post_request_hook = (
            mask_protected_content_post_request_hook(user_context, permissions)
        )
        return True


def _build_nested_matcher(object_lists, keys_info, value, index=0):
    info = keys_info[index]

    if info["object_list"]:
        nested_matcher = _build_nested_matcher(
            object_lists, keys_info, value, index + 1
        )
        elem_match = {
            "$elemMatch": {
                object_lists[info["key"]]: info["object_key"],
                keys_info[index + 1]["key"]: nested_matcher,
            }
        }
        return elem_match if index > 0 else {info["key"]: elem_match}

    if isinstance(value, list):
        value = {"$in": value}
    return value if index > 0 else {info["key"]: value}



from elody.error_codes import ErrorCode, get_error_code, get_read
from configuration import get_object_configuration_mapper  # pyright: ignore
from flask_restful import abort  # pyright: ignore
from serialization.serialize import serialize  # pyright: ignore


def get_content(item, request, content):
    return serialize(
        content,
        type=item.get("type"),
        from_format=serialize.get_format(
            (request.view_args or {}).get("spec", "elody"), request.args
        ),
        to_format=get_object_configuration_mapper().get(item["type"]).SCHEMA_TYPE,
    )


def get_item(storage_manager, user_context_bag, view_args):
    view_args = view_args or {}
    id = view_args.get("id")
    resolve_collections = user_context_bag.get("collection_resolver")
    if not resolve_collections:
        abort(
            403,
            message=f"{get_error_code(ErrorCode.UNDEFINED_COLLECTION_RESOLVER, get_read())} - Collection resolver not defined for user.",
        )
    collections = resolve_collections(collection=view_args.get("collection"), id=id)
    for collection in collections:
        if item := storage_manager.get_db_engine().get_item_from_collection_by_id(
            collection, id
        ):
            return item
    else:
        abort(
            404,
            message=f"{get_error_code(ErrorCode.ITEM_NOT_FOUND, get_read())} | id:{id} - Item with id {id} does not exist.",
        )



import re as regex

from elody.error_codes import ErrorCode, get_error_code, get_read, get_write
from elody.policies.permission_handler import (
    get_permissions,
    handle_single_item_request,
)
from flask import Request  # pyright: ignore
from flask_restful import abort  # pyright: ignore
from inuits_policy_based_auth import BaseAuthorizationPolicy  # pyright: ignore
from inuits_policy_based_auth.contexts.policy_context import (  # pyright: ignore
    PolicyContext,
)
from inuits_policy_based_auth.contexts.user_context import (  # pyright: ignore
    UserContext,
)
from storage.storagemanager import StorageManager  # pyright: ignore


class MediafileDerivativesPolicy(BaseAuthorizationPolicy):
    def authorize(
        self, policy_context: PolicyContext, user_context: UserContext, request_context
    ):
        request: Request = request_context.http_request
        if not regex.match(r"^/mediafiles/(.+)/derivatives$", request.path):
            return policy_context

        view_args = request.view_args or {}
        collection = view_args.get("collection", request.path.split("/")[-3])
        id = view_args.get("id")
        item = (
            StorageManager()
            .get_db_engine()
            .get_item_from_collection_by_id(collection, id)
        )
        if not item:
            abort(
                404,
                message=f"{get_error_code(ErrorCode.ITEM_NOT_FOUND_IN_COLLECTION, get_read())} | id:{id} | collection: {collection} - Item with id {id} doesn't exist in collection {collection}",
            )

        for role in user_context.x_tenant.roles:
            permissions = get_permissions(role, user_context)
            if not permissions:
                continue

            rules = [
                PostRequestRules,
                GetRequestRules,
            ]
            access_verdict = None
            for rule in rules:
                access_verdict = rule().apply(item, user_context, request, permissions)
                if access_verdict != None:
                    policy_context.access_verdict = access_verdict
                    if not policy_context.access_verdict:
                        return policy_context

            if policy_context.access_verdict:
                return policy_context

        return policy_context


class PostRequestRules:
    def apply(
        self, item, user_context: UserContext, request: Request, permissions
    ) -> bool | None:
        if request.method != "POST":
            return None

        return handle_single_item_request(user_context, item, permissions, "create")


class GetRequestRules:
    def apply(
        self, item, user_context: UserContext, request: Request, permissions
    ) -> bool | None:
        if request.method != "GET":
            return None

        return handle_single_item_request(user_context, item, permissions, "read")


class DeleteRequestRules:
    def apply(
        self, item, user_context: UserContext, request: Request, permissions
    ) -> bool | None:
        if request.method != "DELETE":
            return None

        return handle_single_item_request(user_context, item, permissions, "delete")



import json
import mimetypes
import re as regex

from cloudevents.conversion import to_dict
from cloudevents.http import CloudEvent
from collections.abc import MutableMapping, Iterable
from copy import deepcopy
from datetime import datetime, timezone
from os import getenv


URL_UNFRIENDLY_CHARS = {
    " ": "%20",
    "!": "%21",
    "'": "%22",
    "#": "%23",
    "%": "%25",
    "&": "%26",
    '"': "%27",
    "(": "%28",
    ")": "%29",
    "*": "%2A",
    "+": "%2B",
    ",": "%2C",
    "/": "%2F",
    ";": "%3B",
    "<": "%3C",
    "=": "%3D",
    ">": "%3E",
    "?": "%3F",
    "@": "%40",
    "[": "%5B",
    "\\": "%5C",
    "]": "%5D",
    "^": "%5E",
    "`": "%60",
    "{": "%7B",
    "|": "%7C",
    "}": "%7D",
    "~": "%7E",
}


class CustomJSONEncoder(json.JSONEncoder):
    def __convert_datetime(self, obj):
        if obj.tzinfo is None:
            obj = obj.astimezone(timezone.utc)
        return obj.isoformat()

    def default(self, obj):
        if not isinstance(obj, datetime):
            return super().default(obj)
        return self.__convert_datetime(obj)

    def encode(self, obj):
        if not isinstance(obj, datetime):
            return super().encode(obj)
        return self.__convert_datetime(obj)


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def custom_json_dumps(obj):
    return json.dumps(obj, cls=CustomJSONEncoder)


def flatten_dict(object_lists, data: MutableMapping, parent_key=""):
    if not parent_key:
        data = deepcopy(data)
    flat_dict = {}
    for key, value in __flatten_dict_generator(object_lists, data, parent_key):
        if key in flat_dict:
            if not isinstance(flat_dict[key], list):
                flat_dict[key] = [flat_dict[key]]
            if isinstance(value, list):
                flat_dict[key].extend(value)
            else:
                flat_dict[key].append(value)
        else:
            flat_dict[key] = value
    return flat_dict


def __flatten_dict_generator(object_lists, data: MutableMapping, parent_key):
    for key, value in data.items():
        flattened_key = f"{parent_key}.{key}" if parent_key else key
        if isinstance(value, MutableMapping):
            yield from flatten_dict(object_lists, value, flattened_key).items()
        elif isinstance(value, Iterable) and not isinstance(value, (str, bytes)):
            if all(isinstance(item, MutableMapping) for item in value):
                item_key = None
                for item in value:
                    item_key = item.get(object_lists.get(key))
                    if item_key:
                        if isinstance(item, MutableMapping):
                            yield from flatten_dict(
                                object_lists, item, f"{flattened_key}.{item_key}"
                            ).items()
                        else:
                            yield f"{flattened_key}.{item_key}", item["value"]
                if not item_key:
                    yield flattened_key, value
            else:
                yield flattened_key, value
        else:
            yield flattened_key, value


def get_raw_id(item):
    return item.get("_key", item["_id"])


def get_item_metadata_value(item, key):
    for item in item.get("metadata", []):
        if item["key"] == key:
            return item["value"]
    return ""


def get_mimetype_from_filename(filename):
    mime = mimetypes.guess_type(filename, False)[0]
    return mime if mime else "application/octet-stream"


def interpret_flat_key(flat_key: str, object_lists):
    keys_info = []
    index = 0

    flat_key_parts = regex.split(r"\.(?=(?:[^`]*`[^`]*`)*[^`]*$)", flat_key)
    while index < len(flat_key_parts):
        info = {
            "key": flat_key_parts[index],
            "object_list": (
                flat_key_parts[index]
                if flat_key_parts[index] in object_lists.keys()
                else ""
            ),
        }

        if info["object_list"]:
            combined_key = "".join(
                [f"{info['key']}." for info in keys_info if not info["object_list"]]
            )
            info["key"] = f"{combined_key}{info['key']}"
            info["object_key"] = flat_key_parts[index + 1]
            keys_info = [info for info in keys_info if info["object_list"]]

        keys_info.append(info)
        index += 2 if info["object_list"] else 1

    return keys_info


def mediafile_is_public(mediafile):
    publication_status = get_item_metadata_value(mediafile, "publication_status")
    copyright_color = get_item_metadata_value(mediafile, "copyright_color")
    return publication_status.lower() in [
        "beschermd",
        "expliciet",
        "publiek",
    ] or copyright_color.lower() in ["green", "groen"]


def parse_string_to_bool(value):
    if isinstance(value, str):
        value_lower = value.strip().lower()
        if value_lower in ["true", "yes", "1", "y"]:
            return True
        elif value_lower in ["false", "no", "0", "n"]:
            return False
    return value


def read_json_as_dict(filename, logger):
    try:
        with open(filename) as file:
            return json.load(file)
    except (FileNotFoundError, json.JSONDecodeError) as ex:
        logger.error(f"Could not read {filename} as a dict: {ex}")
    return {}


def parse_url_unfriendly_string(
    input: str, *, replace_char=None, return_unfriendly_chars=False
):
    unfriendly_chars = []
    for char, encoded in URL_UNFRIENDLY_CHARS.items():
        parsed_input = input.replace(
            char, encoded if replace_char is None else replace_char
        )
        if input != parsed_input:
            unfriendly_chars.append(char)
        input = parsed_input

    if return_unfriendly_chars:
        return input, unfriendly_chars
    return input


def send_cloudevent(mq_client, source, routing_key, data, exchange_name=None):
    event = to_dict(CloudEvent({"source": source, "type": routing_key}, data))
    if getenv("AMQP_MANAGER", "amqpstorm_flask") in ["amqpstorm_flask"]:
        mq_client.send(event, routing_key=routing_key, exchange_name=exchange_name)
    else:
        mq_client.send(event, routing_key=routing_key)


def signal_child_relation_changed(mq_client, collection, id):
    data = {"parent_id": id, "collection": collection}
    send_cloudevent(mq_client, "dams", "dams.child_relation_changed", data)


def signal_edge_changed(mq_client, parent_ids_from_changed_edges):
    data = {
        "location": f'/entities?ids={",".join(parent_ids_from_changed_edges)}&skip_relations=1'
    }
    send_cloudevent(mq_client, "dams", "dams.edge_changed", data)


def signal_entity_changed(mq_client, entity):
    data = {
        "location": f"/entities/{get_raw_id(entity)}",
        "type": entity.get("type", "unspecified"),
    }
    send_cloudevent(mq_client, "dams", "dams.entity_changed", data)


def signal_entity_deleted(mq_client, entity):
    data = {"_id": get_raw_id(entity), "type": entity.get("type", "unspecified")}
    send_cloudevent(mq_client, "dams", "dams.entity_deleted", data)


def signal_mediafiles_added_for_entity(mq_client, entity, mediafiles):
    data = {"entity": entity, "mediafiles": mediafiles}
    send_cloudevent(mq_client, "dams", "dams.mediafiles_added_for_entity", data)


def signal_relations_deleted_for_entity(mq_client, entity, relations):
    data = {"entity": entity, "relations": relations}
    send_cloudevent(mq_client, "dams", "dams.relations_deleted_for_entity", data)


def signal_mediafile_changed(mq_client, old_mediafile, mediafile):
    data = {"old_mediafile": old_mediafile, "mediafile": mediafile}
    send_cloudevent(mq_client, "dams", "dams.mediafile_changed", data)


def signal_mediafile_deleted(mq_client, mediafile, linked_entities):
    data = {"mediafile": mediafile, "linked_entities": linked_entities}
    send_cloudevent(mq_client, "dams", "dams.mediafile_deleted", data)


def signal_update_copyright_color_entity(mq_client, entity_id):
    data = {"_id": entity_id}
    send_cloudevent(mq_client, "dams", "dams.update_copyright_color_entity", data)


def signal_update_copyright_color_mediafile(mq_client, mediafile_id):
    data = {"_id": mediafile_id}
    send_cloudevent(mq_client, "dams", "dams.update_copyright_color_mediafile", data)


def signal_upload_file(mq_client, upload_links, selected_folder):
    data = {"upload_links": upload_links, "selected_folder": selected_folder}
    send_cloudevent(mq_client, "dams", "dams.upload_file", data)



from enum import Enum


class ErrorCode(Enum):
    READ = "R"
    WRITE = "W"

    # General error codes
    UNKNOWN_ERROR = ("0002", [])
    COLLECTION_NOT_FOUND = ("0003", ["collection"])
    ITEM_NOT_FOUND = ("0004", ["id"])
    HISTORY_ITEM_NOT_FOUND = ("0005", [])
    CANNOT_SPECIFY_BOTH = ("0006", [])
    ALREADY_PARENT = ("0007", ["id"])
    METADATA_KEY_UNDEFINED = ("0008", ["key", "document"])
    ENTITY_HAS_NO_TENANT = ("0009", ["value", "user"])
    MEDIAFILE_NOT_FOUND = ("0010", [])
    ITEM_NOT_FOUND_IN_COLLECTION = ("0011", ["id", "collection"])

    # Auth error codes
    INVALID_CREDENTIALS = ("1001", [])
    ACCOUNT_LOCKED = ("1002", [])
    INSUFFICIENT_PERMISSIONS = ("1003", ["restricted_keys"])
    NO_PERMISSIONS = ("1004", [])
    UNDEFINED_COLLECTION_RESOLVER = ("1005", [])
    TENANT_NOT_FOUND = ("1006", [])
    XTENANT_HAS_NO_TENANT_DEFENING_ENTITY = ("1007", ["x_tenant_id"])
    INSUFFICIENT_PERMISSIONS_WITHOUT_VARS = ("1008", [])
    NO_GLOBAL_ROLES = ("1009", [])
    NO_PERMISSION_TO_TENANT = ("1010", ["tenant_id"])
    XTENANT_NOT_FOUND = ("1011", ["x_tenant_id"])
    NO_DOWNLOAD_PERMISSION = ("1012", [])

    # Database error codes
    DATABASE_NOT_INITIALIZED = ("2000", [])
    DATABASE_CONNECTION_FAILED = ("2001", [])
    QUERY_EXECUTION_FAILED = ("2002", [])
    DUPLICATE_ENTRY = ("2003", [])
    DUPLICATE_IDENTIFIERS = ("2004", ["duplicate_keys"])

    # Network error codes
    NETWORK_UNAVAILABLE = ("3001", [])
    TIMEOUT = ("3002", [])
    SERVER_NOT_FOUND = ("3003", [])

    # File handling error codes
    FILE_NOT_FOUND = ("4001", [])
    FILE_ACCESS_DENIED = ("4002", [])
    FILE_CORRUPTED = ("4003", [])
    NO_FILENAME_SPECIFIED = ("4004", [])
    NO_TICKET_ID_SPECIFIED = ("4005", [])
    TICKET_NOT_FOUND = ("4006", [])
    TICKET_EXPIRED = ("4007", [])
    PROVIDE_MEDIAFILE_ID_OR_TICKET_ID = ("4008", [])
    DUPLICATE_FILE = ("4009", [])
    NO_BUCKET_SPECIFIED = ("4010", [])

    # Validation error codes
    INVALID_INPUT = ("5001", [])
    REQUIRED_FIELD_MISSING = ("5002", [])
    INVALID_FORMAT = ("5003", [])
    INVALID_TYPE = ("5004", [])
    COLUMN_NOT_FOUND = ("5005", ["missing_columns"])
    ONLY_TYPE_CSV_ALLOWED = ("5006", [])
    NO_METADATA_AVAILABLE = ("5007", [])
    INVALID_DATETIME = ("5008", ["value"])
    UNSUPPORTED_TYPE = ("5009", ["type"])
    CONTENT_NOT_FOUND = ("5010", [])
    VALIDATION_ERROR = ("5011", [])
    TENANT_HAS_MISSING_DATA = ("5012", [])
    INVALID_FORMAT_FOR_TYPE = ("5013", ["type"])
    NO_METADATA_AVAILABLE_FOR_ITEM = ("5014", ["id"])
    INVALID_ACCEPT_HEADER = ("5015", [])

    # Filter error codes
    NO_MATCHER_FOR_FILTER_REQUEST = ("6001", [])
    UNDEFINED_FILTER_FOR_INPUT_TYPE = ("6002", ["input_type"])
    UNSUPPORTED_OPERATOR = ("6003", ["operator"])

    # Migration error codes
    UNABLE_TO_UPDATE_SCHEMA_VERSION = ("7001", ["migrated_item"])
    LAZY_MIGRATION_SCHEMA_TYPE_MISMATCH = ("7002", [])

    # External Services
    SERVICE_UNAVAILABLE = ("8001", [])

    # Arches error codes
    ARCHES_ERROR = ("11000", ["error"])
    ARCHES_CONNECTION_UNAVAILABLE = ("11001", [])
    ARCHES_RECORD_NOT_FOUND = ("11002", ["error_message"])
    ARCHES_RECORD_MISSING_DATA = ("11003", ["arches_id"])
    ARCHES_RECORD_MISSING_DATA_DC_PUBLISHER = ("11004", ["arches_id"])
    ARCHES_UNABLE_TO_CREATE_RELATION = ("11005", ["type", "value"])

    # Digipolis error codes
    NO_PERMISSION_TO_CREATE_INSTIUTION = ("12000", ["institution"])
    INSTITUTION_HAS_MISSING_DATA = ("12001", ["institution"])
    INSTITUTION_NOT_FOUND = ("12002", [])


def get_error_code(error_code, prefix):
    if prefix not in [ErrorCode.READ.value, ErrorCode.WRITE.value]:
        raise ValueError("Prefix must be 'R' for read or 'W' for write.")
    return f"{prefix}{error_code.value[0]}"


def get_read():
    return ErrorCode.READ.value


def get_write():
    return ErrorCode.WRITE.value



import re as regex

from elody.error_codes import ErrorCode, get_error_code, get_read, get_write
from flask import Request
from flask_restful import abort
from storage.storagemanager import StorageManager  # pyright: ignore
from elody.util import get_item_metadata_value
from elody.exceptions import NotFoundException, NoTenantException


class TenantIdResolver:
    def resolve(self, request, user):
        endpoints = [
            EntityGetRequest,
            EntityPostRequest,
            EntityDetailGetRequest,
            EntityDetailUpdateRequest,
            EntityDetailDeleteRequest,
            EntityDetailGetRelationsRequest,
            EntityDetailUpdateRelationsRequest,
            EntityDetailCreateRelationsRequest,
            EntityDetailDeleteRelationsRequest,
            EntityDetailGetMetadataRequest,
            EntityDetailUpdateMetadataRequest,
            EntityDetailCreateMetadataRequest,
            EntityDetailGetMediafilesRequest,
            EntityDetailCreateMediafilesRequest,
            MediafileGetRequest,
            MediafilePostRequest,
            MediafileDetailGetRequest,
            MediafileDetailUpdateRequest,
            MediafileDetailDeleteRequest,
            MediafileDetailGetDerivativesRequest,
            MediafileDetailCreateDerivativesRequest,
            MediafileDetailDeleteDerivativesRequest,
        ]

        for endpoint in endpoints:
            tenant_id = endpoint().get_tenant_id(request)
            if tenant_id != None:
                relations = user.get("relations", [])
                if len(relations) < 1:
                    raise NoTenantException(f"User is not attached to a tenant")
                has_tenant_relation = any(
                    relation.get("type") == "hasTenant"
                    and relation.get("key") == tenant_id
                    for relation in relations
                )
                if has_tenant_relation:
                    return tenant_id
                elif len(relations) == 1 and relations[0].get("key") == "tenant:super":
                    return "tenant:super"
                else:
                    return relations[0].get("key")
        return "tenant:super"


class BaseRequest:
    def __init__(self) -> None:
        self.storage = StorageManager().get_db_engine()
        self.super_tenant_id = "tenant:super"
        # TODO refactor this in a more generic way
        self.global_types = [
            "language",
            "type",
            "collectionForm",
            "institution",
            "tag",
            "triple",
            "person",
            "externalRecord",
            "verzameling",
            "arches_record",
            "manufacturer",
            "photographer",
            "creator",
            "assetPart",
            "set",
            "download",
            "license",
            "share_link"
            # TODO Mediafile should have a link to an asset
            "mediafile",
            "savedSearch",
            "original_data",
            "user",
        ]

    def _get_tenant_id_from_entity(self, entity_id):
        entity_relations = self.storage.get_collection_item_relations(
            "entities", entity_id
        )
        entity = self.storage.get_item_from_collection_by_id("entities", entity_id)
        if not entity:
            abort(
                404,
                message=f"{get_error_code(ErrorCode.ITEM_NOT_FOUND, get_read())} | id:{entity_id} - Item with id {entity_id} doesn't exist",
            )
        type = entity.get("type")
        if type in self.global_types:
            return "tenant:super"
        if entity_relations:
            for entity_relation in entity_relations:
                if entity_relation.get("type") == "isIn":
                    tenant = self.storage.get_item_from_collection_by_id(
                        "entities", entity_relation.get("key")
                    )
                    return tenant["_id"]
                if entity_relation.get("type") == "hasInstitution":
                    instution_relations = self.storage.get_collection_item_relations(
                        "entities", entity_relation.get("key")
                    )
                    for institution_relation in instution_relations:
                        if institution_relation.get("type") == "defines":
                            return institution_relation.get("key")

        if entity and get_item_metadata_value(entity, "institution"):
            return f"tenant:{get_item_metadata_value(entity, 'institution')}"
        abort(
            400,
            message=f"{get_error_code(ErrorCode.ENTITY_HAS_NO_TENANT, get_read())} - Entity has no tenant, and is suppose to have one.",
        )

    def _get_tenant_id_from_mediafile(self, mediafile_id):
        mediafile_relations = self.storage.get_collection_item_relations(
            "mediafiles", mediafile_id
        )

        for relation in mediafile_relations:
            if relation.get("type") == "belongsTo":
                return self._get_tenant_id_from_entity(relation.get("key"))

    # TODO Mediafile should have a link to an asset
    def _get_tenant_id_from_body(self, item, soft_call=False):
        if soft_call:
            return "tenant:super"
        if item.get("type") in self.global_types or item.get("type") == "institution":
            return "tenant:super"
        if item.get("type") == "asset":
            for relation in item.get("relations", []):
                if relation.get("type") in ["hasArchesLink", "hasAdlibLink", "hasBrocadeLink"]:
                    return "tenant:super"
        institution_id = get_item_metadata_value(item, "institution")
        if not institution_id:
            for relation in item.get("relations", []):
                if relation.get("type") == "hasInstitution":
                    institution_id = relation.get("key")
        if institution_id:
            return f"tenant:{institution_id}"
        abort(
            400,
            message=f"{get_error_code(ErrorCode.ENTITY_HAS_NO_TENANT, get_read())} - Item in body doesn't have an institution.",
        )


class EntityGetRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/entities(?:\?(.*))?$", request.path)
            and request.method == "GET"
        ):
            return self.super_tenant_id
        return None


class EntityPostRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/entities(?:\?(.*))?$", request.path)
            and request.method == "POST"
        ):
            is_soft_call = request.args.get("soft") is not None
            return self._get_tenant_id_from_body(request.json, soft_call=is_soft_call)
        return None


class EntityDetailGetRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if regex.match(r"/entities/(.+)$", request.path) and request.method == "GET":
            return self._get_tenant_id_from_entity(request.view_args.get("id"))
        return None


class EntityDetailUpdateRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if regex.match(r"^/entities/(.+)$", request.path) and request.method in [
            "PUT",
            "PATCH",
        ]:
            return self._get_tenant_id_from_entity(request.view_args.get("id"))
        return None


class EntityDetailDeleteRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/entities/([^/]+)$", request.path)
            and request.method == "DELETE"
        ):
            return self._get_tenant_id_from_entity(request.view_args.get("id"))
        return None


class EntityDetailGetRelationsRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/entities/(.+)/relations$", request.path)
            and request.method == "GET"
        ):
            return self._get_tenant_id_from_entity(request.view_args.get("id"))
        return None


class EntityDetailUpdateRelationsRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if regex.match(
            r"^/entities/(.+)/relations$", request.path
        ) and request.method in ["PUT", "PATCH"]:
            return self._get_tenant_id_from_entity(request.view_args.get("id"))
        return None


class EntityDetailCreateRelationsRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/entities/(.+)/relations$", request.path)
            and request.method == "POST"
        ):
            return self._get_tenant_id_from_entity(request.view_args.get("id"))
        return None


class EntityDetailDeleteRelationsRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/entities/(.+)/relations$", request.path)
            and request.method == "DELETE"
        ):
            return self._get_tenant_id_from_entity(request.view_args.get("id"))
        return None


class EntityDetailGetMetadataRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/entities/(.+)/metadata$", request.path)
            and request.method == "GET"
        ):
            return self._get_tenant_id_from_entity(request.view_args.get("id"))
        return None


class EntityDetailUpdateMetadataRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if regex.match(
            r"^/entities/(.+)/metadata$", request.path
        ) and request.method in ["PUT", "PATCH"]:
            return self._get_tenant_id_from_entity(request.view_args.get("id"))
        return None


class EntityDetailCreateMetadataRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/entities/(.+)/metadata$", request.path)
            and request.method == "POST"
        ):
            return self._get_tenant_id_from_entity(request.view_args.get("id"))
        return None


class EntityDetailGetMediafilesRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/entities/(.+)/mediafiles$", request.path)
            and request.method == "GET"
        ):
            return self._get_tenant_id_from_entity(request.view_args.get("id"))
        return None


class EntityDetailCreateMediafilesRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/entities/(.+)/mediafiles$", request.path)
            and request.method == "POST"
        ):
            return self._get_tenant_id_from_entity(request.view_args.get("id"))
        return None


# Mediafiles
class MediafileGetRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/mediafiles(?:\?(.*))?$", request.path)
            and request.method == "GET"
        ):
            return self.super_tenant_id
        return None


class MediafilePostRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/mediafiles(?:\?(.*))?$", request.path)
            and request.method == "POST"
        ):
            is_soft_call = request.args.get("soft") is not None
            return self._get_tenant_id_from_body(request.json, soft_call=is_soft_call)
        return None


class MediafileDetailGetRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/mediafiles/([^/]+)$", request.path)
            and request.method == "GET"
        ):
            return self.super_tenant_id
        return None


class MediafileDetailUpdateRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if regex.match(r"^/mediafiles/(.+)$", request.path) and request.method in [
            "PUT",
            "PATCH",
        ]:
            return self._get_tenant_id_from_mediafile(request.view_args.get("id"))
        return None


class MediafileDetailDeleteRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/mediafiles/(.+)$", request.path)
            and request.method == "DELETE"
        ):
            return self._get_tenant_id_from_mediafile(request.view_args.get("id"))
        return None


class MediafileDetailGetDerivativesRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/mediafiles/(.+)/derivatives$", request.path)
            and request.method == "GET"
        ):
            return self.super_tenant_id
        return None


class MediafileDetailCreateDerivativesRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/mediafiles/(.+)/derivatives$", request.path)
            and request.method == "POST"
        ):
            return self._get_tenant_id_from_mediafile(request.view_args.get("id"))
        return None


class MediafileDetailDeleteDerivativesRequest(BaseRequest):
    def get_tenant_id(self, request: Request) -> str | None:
        if (
            regex.match(r"^/mediafiles/(.+)/derivatives$", request.path)
            and request.method == "DELETE"
        ):
            return self._get_tenant_id_from_mediafile(request.view_args.get("id"))
        return None
